/**
 * Metrics returned by the `runQueries` function. These contain various
 * performance numbers and a unique identifier which can be used to trace
 * the request in Sentry.
 */
interface Metrics {
    /**
     * The Trace ID provided by Sentry. This is only available when the required
     * Sentry environment variables are present.
     */
    id: string | null;
    /** The total duration it took the worker to process the request. */
    duration: number;
}

type RecursivePartial<T> = {
    [K in keyof T]?: T[K] extends Array<infer U> ? Array<RecursivePartial<U>> : T[K] extends object ? RecursivePartial<T[K]> : T[K];
};
type QueryInstructions = {
    with: Record<string, number | string | null | any>;
    where: any;
    to: Record<string, number | string | null | any>;
    including: string[];
    excluding: string[];
    orderedBy: {
        ascending: string[];
        descending: string[];
    };
    in: string;
    after: number;
    limitedTo: number;
};
type QueryType = 'get' | 'set' | 'drop' | 'create' | 'count';
type QuerySchema = Record<string, QueryInstructions>;
type Query = Record<QueryType, QuerySchema>;
interface QueryListOptions {
    moreAfter?: number;
}
type QueryResult<T> = T extends [] ? [] : T extends [infer First, ...infer Rest] ? Rest extends unknown ? First extends Promise<infer U> ? [U] : BindToArray<[First, ...Rest], QueryListOptions> : Array<First> : T extends Promise<infer U> ? [U] : [T];
type ReadableValue = File | ReadableStream;
type Reference = {
    key: string;
} | Media;
interface Readable {
    /** Index and type of the Query the `Readable` was found in. */
    query: {
        index: number;
        type: 'set' | 'create';
    };
    /** Schema to update. */
    schema: string;
    /** Name of the field to set the value to. */
    field: string;
    /** Stream containing the value to be set. */
    value: ReadableValue;
    /** Content-Type of the streamed value. */
    contentType: string;
}
interface StoredReadable extends Readable {
    /** Reference to the stored value or Media object. */
    reference: Reference;
}
type MediaMetadata = {
    /** Size of the image or video in bytes. */
    size: number | null;
    /** MIME type of the image or video. */
    type: string | null;
    /** Width of the image or video in pixels. */
    width: number | null;
    /** Height of the image or video in pixels. */
    height: number | null;
};
type MediaPlaceholder = {
    /** Base64 representation of the generated placeholder. */
    base64: string | null;
};
type StoredMedia = {
    /** Key under which the media item is stored inside RONIN. */
    key: string;
    /** Object containing various metadata of the stored media item. */
    meta: MediaMetadata;
    /** Object containing different versions of the generated placeholder. */
    placeholder: MediaPlaceholder | null;
};
type Media = StoredMedia & {
    /** URL under which the stored image or video can be accessed from the client.  */
    src: string;
};
/**
 * @see https://twitter.com/mattpocockuk/status/1622730173446557697
 */
type Prettify<T> = T extends Array<unknown> ? T : {
    [K in keyof T]: T[K];
} & {};
type BindToArray<T, R> = {
    [K in keyof T]: T[K] extends Array<infer U> ? Array<U> & R : T[K];
};

type Fetcher = typeof fetch;

type ReadableProcessingOptions = Partial<{
    /**
     * Allows for specifying a custom endpoint `Readable`s will be uploaded to.
     *
     * This endpoint has to respond with a JSON object of the following format:
     *
     * ```js
     * const res = (...) => {
     *   // Where the value of `id` has to be a valid UUID v4.
     *   key: "35bc40a1-eaec-4163-8259-37c6b9723740"
     * }
     * ```
     */
    endpoint: string;
    /** Allows for specifying a custom HTTP request method. Defaults to `PUT`. */
    method: string;
    /** Allows for specifying a custom `fetch` function used to make network requests. */
    fetcher?: Fetcher;
}>;
/**
 * ## 🚧 For internal use only! 🚧
 *
 * Store `Readable`s from given queries and populate the queries with the respective references.
 * @param queries `Query` array which may or may not include `Readable`s.
 * @param options `ReadableProcessingOptions` object containing options passed on to the internal handling of `Readable`s.
 * @returns Queries populated with the references to each `Readable` they might have initially contained.
 */
declare const processReadables: (queries: Query[], token: string, options?: ReadableProcessingOptions) => Promise<Query[]>;

type QueryProcessingOptions = {
    /**
     * Allows for specifying a custom HTTP endpoint that the provided queries
     * should be submitted to.
     */
    endpoint?: string;
    /**
     * Allows for specifying a custom replacement for the native JavaScript
     * `fetch` function used to make network requests for submitting queries.
     */
    fetcher?: Fetcher;
    /** ## 🚧 For internal use only! 🚧 */
    dataSelector?: string;
};

declare namespace RONIN {
    export type RecursiveNullish<T> = T extends Array<any> ? {
        [K in keyof T]?: RecursiveNullish<T[K]> | null;
    } | null | undefined : T extends Date ? T | null | undefined : T extends Record<string, any> | RONIN.RoninRecord ? {
        [K in keyof T]?: RecursiveNullish<T[K]> | null;
    } : T | null | undefined;
    export interface RoninMetadata {
        createdAt: Date;
        createdBy: `acc_${string}` | Record<string, any>;
        locked: boolean;
        status: 'draft' | 'published' | 'archived';
        updatedAt: Date;
        updatedBy: `acc_${string}` | Record<string, any>;
    }
    export interface RoninRecord<TIdPrefix extends string = 'rec'> {
        id: `${TIdPrefix}_${string}`;
        ronin: RoninMetadata;
    }
    interface AllSubFilters<T> {
        is: T | Array<T> | null;
        isNot: T | Array<T> | null;
        contains: string | Exclude<T | Array<T>, null>;
        endsWith: string | Exclude<T | Array<T>, null>;
        startsWith: string | Exclude<T | Array<T>, null>;
        isGreaterThan: number | Date;
        isLessThan: number | Date;
    }
    type SubFilter<T> = Prettify<T extends number ? Pick<Partial<AllSubFilters<T>>, 'isGreaterThan' | 'isLessThan' | 'is'> : T extends string ? Pick<Partial<AllSubFilters<T>>, 'contains' | 'endsWith' | 'is' | 'isNot' | 'startsWith'> : T extends boolean ? Pick<Partial<AllSubFilters<T>>, 'is'> : T extends Date ? Pick<Partial<AllSubFilters<T>>, 'isGreaterThan' | 'isLessThan' | 'is'> : never>;
    type Where<TSchema extends Record<string, any>> = {
        [TKey in keyof TSchema]: SubFilter<TSchema[TKey]> | Array<SubFilter<TSchema[TKey]>>;
    };
    type With<TSchema extends Record<string, any>> = {
        [TKey in keyof TSchema]: TSchema[TKey] | Array<TSchema[TKey]>;
    };
    type AllFields<TSchema> = `ronin.${keyof RoninMetadata}` | Exclude<keyof TSchema, 'ronin'>;
    interface OrderedBy<TSchema> {
        /** Order the resulting records in descending order using a specific field. */
        descending: Array<AllFields<TSchema>>;
        /** Order the resulting records in ascending order using a specific field. */
        ascending: Array<AllFields<TSchema>>;
    }
    type To<TSchema extends Record<string, any>> = With<TSchema>;
    export type SchemaSlugKey = keyof RONIN.Counter | keyof RONIN.Creator | keyof RONIN.Dropper | keyof RONIN.Getter | keyof RONIN.Setter;
    export interface Filter<TSchema extends (Record<string, any> & RoninRecord) | Array<Record<string, any> & RoninRecord>, TSlug extends SchemaSlugKey, TVariant extends string = never> {
        /** Perform specific matching actions (such as "is", "contains", or "startsWith") on the fields of a record. */
        where: RecursiveNullish<Where<TSchema>> | Array<RecursiveNullish<Where<TSchema>>>;
        /** Perform "is" matching actions on the fields of a record. */
        with: RecursiveNullish<With<TSchema>> | Array<RecursiveNullish<With<TSchema>>>;
        /** Order the resulting records using a specific field. */
        orderedBy: RecursivePartial<OrderedBy<TSchema>>;
        /** Resolve records in a specific variant. */
        in: TVariant;
        /** Resolve the records associated with certain fields of type "Record". */
        including: Array<RecursivePartial<RONIN.Includes[TSlug]> | `ronin.${keyof Pick<RoninMetadata, 'createdBy' | 'updatedBy'>}`>;
        /**
         * Set the maximum number of records to be retrieved before a subsequent page must be requested.
         *
         * The default is `20` and the maximum is `1000`.
         */
        limitedTo: number;
        /**
         * Retrieve the next page of records.
         *
         * Consumes the value of `records.moreAfter` (defined on the array of records returned by a previous query).
         *
         * The page size can be customized using `limitedTo` and defaults to `20`.
         */
        after: number;
    }
    export interface FilterAndMutate<TSchema extends (Record<string, any> & RoninRecord) | Array<Record<string, any> & RoninRecord>, TSlug extends SchemaSlugKey, TVariant extends string = never> extends Filter<TSchema, TSlug, TVariant> {
        /** Perform mutations on the fields of a record or add new fields to it. */
        to: RecursiveNullish<To<TSchema>>;
    }
    export interface Includes extends Record<SchemaSlugKey, string> {
    }
    export interface Counter extends Record<string, Partial<Filter<Record<string, any> & RoninRecord, never>>> {
    }
    export interface Creator extends Record<string, Partial<Filter<Record<string, any> & RoninRecord, never>>> {
    }
    export interface Dropper extends Record<string, Partial<Filter<Record<string, any> & RoninRecord, never>>> {
    }
    export interface Getter extends Record<string, Partial<Filter<Record<string, any> & RoninRecord, never>>> {
    }
    export interface Setter extends Record<string, Partial<FilterAndMutate<Record<string, any> & RoninRecord, never>>> {
    }
    export {};
}

interface RunScratchpadOptions {
    /**
     * Allows for specifying custom data hooks that are executed for every query
     * of a particular schema.
     */
    hooks?: Record<string, unknown>;
    /**
     * Allows for passing a custom app token. In a Node.js-based environment,
     * `process.env.RONIN_TOKEN` will be used by default.
     */
    token?: string;
    /**
     * Allows for extending the lifetime of the edge worker invocation until the
     * provided promise has been resolved.
     */
    waitUntil?: (promise: Promise<unknown>) => void;
    /**
     * ## 🚧 For internal use only! 🚧
     *
     * Allows for specifying custom options used for processing readables.
     */
    readableProcessingOptions?: ReadableProcessingOptions;
    /**
     * ## 🚧 For internal use only! 🚧
     *
     * Allows for specifying custom options used for processing queries.
     */
    queryProcessingOptions?: QueryProcessingOptions;
}
interface ScratchPadActions {
    /**
     * Retrieve existing records.
     *
     * Read more at https://docs.ronin.co/queries
     *
     * ```js
     * // Simple Example
     * get.account.with.handle = "juri";
     *
     * // Advanced Example
     * get.accounts = {
     *   where: {
     *     role: {
     *       is: "ADMIN" // 👈🏻 Match specific fields
     *     }
     *   }
     *   orderedBy: {
     *     ascending: ["createdAt"]
     *   },
     *   include: ["roles"], // 👈🏻 Resolve nested records
     * }
     * ```
     */
    get: RONIN.Getter;
    /**
     * Modify existing records.
     *
     * Read more at https://docs.ronin.co/queries
     *
     * ```js
     * // Example
     * set.account = {
     *   with: {
     *     handle: "juri",
     *   },
     *   to: {
     *     verified: true,
     *   },
     * }
     * ```
     */
    set: RONIN.Setter;
    /**
     * Create a new record.
     *
     * Read more at https://docs.ronin.co/queries
     *
     * ```js
     * // Example
     * create.account.with = {
     *   handle: "juri",
     *   verified: false,
     * }
     * ```
     */
    create: RONIN.Creator;
    /**
     * Drop existing records.
     *
     * Read more at https://docs.ronin.co/queries
     *
     * ```js
     * // Example
     * drop.accounts.where.handle.isNot = "juri",
     * ```
     */
    drop: RONIN.Dropper;
    /**
     * Count matching records.
     *
     * Read more at https://docs.ronin.co/queries
     *
     * ```js
     * // Example:
     * count.accounts.where.verified.is = true;
     * ```
     */
    count: RONIN.Counter;
}
type ScratchPad = (actions: ScratchPadActions) => void;
type ScratchPadFactory = <T>(scratchPad: ScratchPad) => Promise<QueryResultWithMetrics<T>>;
type QueryResultWithMetrics<T> = QueryResult<T> & {
    metrics?: Metrics;
};

type FilteredHookQuery<TQuery extends QueryInstructions, TType extends QueryType> = RecursivePartial<TQuery> & Pick<TQuery, TType extends 'count' ? never : TType extends 'create' ? 'with' : TType extends 'get' ? never : TType extends 'set' ? 'to' : TType extends 'drop' ? never : never>;
type BeforeHookHandler<TType extends QueryType, TQuery extends FilteredHookQuery<QueryInstructions, TType> = FilteredHookQuery<QueryInstructions, TType>> = (query: TQuery, multipleRecords: boolean) => TQuery | Promise<TQuery>;
type DuringHookHandler<TType extends QueryType, TSchema = unknown> = (query: FilteredHookQuery<QueryInstructions, TType>, multipleRecords: boolean) => TSchema | Promise<TSchema>;
type AfterHookHandler<TType extends QueryType, TSchema = unknown> = (query: FilteredHookQuery<QueryInstructions, TType>, multipleRecords: boolean, result: TSchema) => void | Promise<void>;
type HookType = 'before' | 'during' | 'after';
type Hook<TStage extends HookType, TType extends QueryType, TSchema extends TStage extends 'before' ? never : unknown = never> = TStage extends 'before' ? BeforeHookHandler<TType> : TStage extends 'during' ? DuringHookHandler<TType, TSchema> : TStage extends 'after' ? AfterHookHandler<TType, TSchema> : never;
type BeforeHook<TType extends QueryType> = Hook<'before', TType>;
type DuringHook<TType extends QueryType, TSchema = unknown> = Hook<'during', TType, TSchema>;
type AfterHook<TType extends QueryType, TSchema = unknown> = Hook<'after', TType, TSchema>;
type BeforeCountHook = BeforeHook<'count'>;
type BeforeCreateHook = BeforeHook<'create'>;
type BeforeDropHook = BeforeHook<'drop'>;
type BeforeGetHook = BeforeHook<'get'>;
type BeforeSetHook = BeforeHook<'set'>;
type CountHook<TSchema = unknown> = DuringHook<'count', TSchema>;
type CreateHook<TSchema = unknown> = DuringHook<'create', TSchema>;
type DropHook<TSchema = unknown> = DuringHook<'drop', TSchema>;
type GetHook<TSchema = unknown> = DuringHook<'get', TSchema>;
type SetHook<TSchema = unknown> = DuringHook<'set', TSchema>;
type AfterCountHook<TSchema = unknown> = AfterHook<'count', TSchema>;
type AfterCreateHook<TSchema = unknown> = AfterHook<'create', TSchema>;
type AfterDropHook<TSchema = unknown> = AfterHook<'drop', TSchema>;
type AfterGetHook<TSchema = unknown> = AfterHook<'get', TSchema>;
type AfterSetHook<TSchema = unknown> = AfterHook<'set', TSchema>;

/**
 * Transform a given `ScratchPad` into `Queries` (array) which can be passed to RONIN.
 * @param scratchPad `ScratchPad` to get executable `Queries` from.
 * @returns `Queries` which will be passed to RONIN.
 */
declare const getQueriesFromScratchPad: (scratchPad: ScratchPad) => Query[];

declare const runQueries: <T>(queries: Query[], options?: RunScratchpadOptions) => Promise<QueryResultWithMetrics<T>>;

/**
 * Execute queries on RONIN.
 *
 * @param config If a scratchpad is passed as the first argument, this can be optional config for it.
 *
 * @returns A function for running a scratchpad.
 */
declare function handleScratchPadOrConfig(config?: RunScratchpadOptions): ScratchPadFactory;
/**
 * Execute queries on RONIN.
 *
 * @param input A scratchpad containing queries, or config for creating a factory for running scratchpads.
 * @param config If a scratchpad is passed as the first argument, this can be optional config for it.
 *
 * @returns The results of the queries in the scratchpad.
 */
declare function handleScratchPadOrConfig<T = [unknown]>(scratchPad: ScratchPad, config?: RunScratchpadOptions): Promise<QueryResult<T>>;

export { AfterCountHook, AfterCreateHook, AfterDropHook, AfterGetHook, AfterHookHandler, AfterSetHook, BeforeCountHook, BeforeCreateHook, BeforeDropHook, BeforeGetHook, BeforeHookHandler, BeforeSetHook, CountHook, CreateHook, DropHook, DuringHookHandler, GetHook, Metrics, Query, QueryInstructions, QueryResult, QueryType, RONIN, Readable, ReadableValue, RunScratchpadOptions, ScratchPad, ScratchPadActions, SetHook, StoredReadable, handleScratchPadOrConfig as default, getQueriesFromScratchPad, processReadables, runQueries };
